#!/usr/bin/python
# -*- coding: ascii -*-
###########################################################################
# simple-gecrypt-decrypt.py - Simplified implementation of gecrypt decryption
#
# Copyright (C) 2007 Dwayne C. Litzenberger <dlitz@dlitz.net>
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation.
# 
# THE AUTHOR PROVIDES THIS SOFTWARE ``AS IS'' AND ANY EXPRESSED OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Country of origin: Canada
#
###########################################################################
# Version 0.3 - 2007-07-31
###########################################################################

import sys
import os.path
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))   # PBKDF2.py is in the parent directory

from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA256
from os import urandom
from PBKDF2 import PBKDF2
from binascii import a2b_hex
from struct import pack, unpack
from CTRCipher import CTRCipher

# Definitions

def abort(error):
    print >>sys.stderr, "error: " + error
    sys.exit(1)

def read_bytes(bytes, file=sys.stdin):
    data = file.read(bytes)
    if len(data) != bytes:
        abort("premature end of file")
    return data

def read_and_hash(digest, bytes, file=sys.stdin):
    data = read_bytes(bytes, file)
    digest.update(data)
    return data

def AES_CTR(key, nonce, message):
    return CTRCipher(key, nonce, AES).encrypt(message)

def HMAC_SHA256(key, message):
    return HMAC.new(key, message, digestmod=SHA256).digest()

FILE_ID     = a2b_hex("616d1d67ca294e2eb98bc01ff0470200")
HEADER_ID   = a2b_hex("616d1d67ca294e2eb98bc01ff0470201")

Int8 = lambda n: chr(n)
Int16 = lambda n: pack("!H", n)
Int64 = lambda n: pack("!Q", n)

Unpack8 = lambda s: ord(s)
Unpack16 = lambda s: unpack("!H", s)[0]
Unpack64 = lambda s: unpack("!Q", s)[0]

# Get input from the user
if sys.argv[1:]:
    keyfile = sys.argv[1]
else:
    print >>sys.stderr, "Usage: simple-gecrypt-decrypt.py keyfile < inputfile > outputfile"
    sys.exit(1)

if sys.argv[2:]:
    c = int(sys.argv[2])
    if c < 10 or c > 0xffff:
        raise ValueError("iteration count must be between 10 and 65535")
else:
    c = 1000

if sys.argv[3:]:
    p = int(sys.argv[3])
    if p < 0:
        raise ValueError("padsize must be at least 0")
else:
    p = 0

# Read master key
K_M = open(keyfile, "rb").read()

# Read FILE_ID
id = read_bytes(16)
if id != FILE_ID:
    abort("unrecognized file format")

# Read salt
S = read_bytes(16)

# Read header nonce
N_H = read_bytes(8)

# Read iteration count
c = Unpack16(read_bytes(2))

# Compute the authentication and cipher keys
K = PBKDF2(K_M, FILE_ID + S, c, digestmodule=SHA256).read(96)
K_A = K[:64]
K_H = K[64:]

# Start hashing
h = HMAC.new(K_A, digestmod=SHA256)
h.update(id + S + N_H + Int16(c))

# Start decrypting the header
hcipher = CTRCipher(K_H, N_H, AES)

# Read the encrypted portion of the header
hid = hcipher.decrypt(read_and_hash(h, 16))
if hid != HEADER_ID:
    abort("unable to decrypt")
N_P = hcipher.decrypt(read_and_hash(h, 8))
key_size = Unpack8(hcipher.decrypt(read_and_hash(h, 1)))
if key_size not in (16, 24, 32):
    abort("unable to decrypt")
K_P = hcipher.decrypt(read_and_hash(h, key_size))

# Authenticate the header
real_mac = h.digest()
mac = read_and_hash(h, 32)  # SHA-256 hash
if real_mac != mac:
    abort("header authentication failed")

# Start decrypting the payload
cipher = CTRCipher(K_P, N_P, AES)

# Read the payload data
for is_padding in (False, True):
    PTXT = []
    L_i = Unpack16(cipher.decrypt(read_and_hash(h, 2)))
    while L_i > 0:
        M_i = cipher.decrypt(read_and_hash(h, L_i))
        
        # Authenticate
        real_mac = h.digest()
        mac = read_and_hash(h, 32)  # SHA-256 hash
        if real_mac != mac:
            abort("payload authentication failed")
        
        PTXT.append(M_i)
        L_i = Unpack16(cipher.decrypt(read_and_hash(h, 2)))
    
    # Authenticate
    real_mac = h.digest()
    mac = read_and_hash(h, 32)  # SHA-256 hash
    if real_mac != mac:
        abort("payload authentication failed")
    
    if not is_padding:
        PLAINTEXT = "".join(PTXT)  # convert list to string

# Output the plaintext
sys.stdout.write(PLAINTEXT)
