#!/usr/bin/python
# -*- coding: ascii -*-
###########################################################################
# simple-gecrypt-encrypt.py - Simplified implementation of gecrypt encryption
#
# Copyright (C) 2007 Dwayne C. Litzenberger <dlitz@dlitz.net>
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation.
# 
# THE AUTHOR PROVIDES THIS SOFTWARE ``AS IS'' AND ANY EXPRESSED OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Country of origin: Canada
#
###########################################################################
# Version 0.3 - 2007-07-31
###########################################################################

import sys
import os.path
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))   # PBKDF2.py is in the parent directory

from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA256
from os import urandom
from PBKDF2 import PBKDF2
from binascii import a2b_hex
from struct import pack
from CTRCipher import CTRCipher

# Definitions

def AES_CTR(key, nonce, message):
    return CTRCipher(key, nonce, AES).encrypt(message)

def HMAC_SHA256(key, message):
    return HMAC.new(key, message, digestmod=SHA256).digest()

FILE_ID     = a2b_hex("616d1d67ca294e2eb98bc01ff0470200")
HEADER_ID   = a2b_hex("616d1d67ca294e2eb98bc01ff0470201")

Int8 = lambda n: chr(n)
Int16 = lambda n: pack("!H", n)
Int64 = lambda n: pack("!Q", n)

# Get input from the user
if sys.argv[1:]:
    keyfile = sys.argv[1]
else:
    print >>sys.stderr, "Usage: simple-gecrypt-encrypt.py keyfile [iteration-count] [pad-size] < inputfile > outputfile"
    sys.exit(1)

if sys.argv[2:]:
    c = int(sys.argv[2])
    if c < 10 or c > 0xffff:
        raise ValueError("iteration count must be between 10 and 65535")
else:
    c = 1000

if sys.argv[3:]:
    p = int(sys.argv[3])
    if p < 0:
        raise ValueError("padsize must be at least 0")
else:
    p = 0

# Read master key
K_M = open(keyfile, "rb").read()

# Read plaintext
M = sys.stdin.read()

# 2. Randomly generate:
#   - 128-bit salt: S
#   - 64-bit nonces: N_H, N_P
#   - 128-, 192-, or 256-bit payload cipher key: K_P
S = urandom(16)
N_H = urandom(8)
N_P = urandom(8)
K_P = urandom(24)

# 3. Using the password-based key derivation function PBKDF2(P, S, c, dkLen)
# from RSA PKCS#5 v2.0, calculate the session key (using HMAC-SHA256 as the
# underlying pseudorandom function)
K = PBKDF2(K_M, FILE_ID + S, c, digestmodule=SHA256).read(96)

# 4. Set the authentication key K_A to be the first 512 bits of the session key.
K_A = K[:64]

# 5. Set the header cipher key K_H to be the last 256 bits of the session key.
K_H = K[64:]

# 6. Generate encrypted portion of the header
C_H = AES_CTR(K_H, N_H, HEADER_ID + N_P + Int8(len(K_P)) + K_P)

# 7. Generate the header
HEADER = FILE_ID + S + N_H + Int16(c) + C_H

# 8. Initialize the payload cipher.
E = CTRCipher(K_P, N_P, AES).encrypt

# 9. Split the plaintext into n chunks M_1, M_2, ... M_n such that the length
# (in octets) of each chunk falls within the interval [1..2^16-1].
M_ = {}
(n, remainder) = divmod(len(M), 0xffff)
if remainder > 0:
    n += 1
del remainder
for i in range(n):
    M_[i+1] = M[i*0xffff:(i+1)*0xffff]

# 10. Generate the padding.
PAD = "\0" * p

# 11. Split the padding into m chunks P_1, P_2, ... P_m such that the length
# (in octets) of each chunk falls within the interval [1..2^16-1].
P_ = {}
(m, remainder) = divmod(len(PAD), 0xffff)
if remainder > 0:
    m += 1
del remainder
for i in range(m):
    P_[i+1] = PAD[i*0xffff:(i+1)*0xffff]

# 12. Generate the ciphertext:
C_ = {}
A_ = {}
L_ = {}
LP_ = {}
CP_ = {}
AP_ = {}

C_[0] = HEADER
A_[0] = HMAC_SHA256(K_A, C_[0])

for i in range(1, n+1):
    L_[i] = Int16(len(M_[i]))
    C_[i] = C_[i-1] + A_[i-1] + E(L_[i] + M_[i])
    A_[i] = HMAC_SHA256(K_A, C_[i])

LP_[0] = Int16(0)
CP_[0] = C_[n] + A_[n] + E(LP_[0])
AP_[0] = HMAC_SHA256(K_A, CP_[0])

for i in range(1, m+1):
    LP_[i] = Int16(len(P_[i]))
    CP_[i] = CP_[i-1] + AP_[i-1] + E(LP_[i] + P_[i])
    AP_[i] = HMAC_SHA256(K_A, CP_[i])

LP_[m+1] = Int16(0)
CP_[m+1] = CP_[m] + AP_[m] + E(LP_[m+1])
AP_[m+1] = HMAC_SHA256(K_A, CP_[m+1])

CIPHERTEXT = CP_[m+1] + AP_[m+1]

# 13. Output CIPHERTEXT.
sys.stdout.write(CIPHERTEXT)

# Done.
